/*
** Module   : AbstractTempTable.java
** Abstract : Abstract base class for static and dynamic temp-table 4GL objects.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20131115 Created initial version.
** 002 CA  20131209 The dynamic flag must be passed up the c'tor call chain.
** 003 HC  20131215 Implemented ADM-DATA and UNIQUE-ID attribute.
** 004 SVL 20140106 ERROR-STRING attribute and some error handling functions are now handled by
**                  this class.
** 005 SVL 20140124 Some functionality was moved here from the subclasses. Implemented
*                   hasRecords().
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;
import org.hibernate.*;
import java.util.*;

/**
 * Abstract base class for static and dynamic temp-table 4GL objects.
 */
public abstract class AbstractTempTable
extends HandleChain
implements TempTable,
           Deletable
{
   /** Default buffer handle. */
   protected handle defaultBufferHandle = null;

   /** Corresponds ERROR-STRING attribute. **/
   protected final character errorString = new character();

   /** Corresponds ERROR attribute. **/
   protected final logical error = new logical(false);

   /** Corresponds ADM-DATA attribute. **/
   protected final character admData = new character();

   /** Corresponds UNIQUE-ID attribute. **/
   private final integer uniqueId = new integer();

   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or a dynamic temp-table.
    */
   public AbstractTempTable(boolean dynamic)
   {
      super(dynamic);
      uniqueId.assign(UniqueIdGenerator.getUniqueId(UniqueIdGenerator.IdKind.TEMP_TABLE));
   }

   /**
    * Gets the the unique ID number associated to this object by the
    * underlying system. Not the same as the handle value itself.
    *
    * @return   The integer number of the children.
    */
   @Override
   public integer getUniqueID()
   {
      if (uniqueId == null)
      {
         // not yet initialized, can occur at construction time
         return new integer();
      }

      return new integer(this.uniqueId);
   }

   /**
    * Sets the ERROR-STRING attribute of 4GL temp-table.
    *
    * @param   errStr
    *          The new value for the ERROR-STRING attribute.
    */
   public void changeErrorString(character errStr)
   {
      if (!prepared().getValue())
      {
         ErrorManager.displayError(12346,
               "ERROR-STRING can only be used with a PREPARED temp-table that is not a BEFORE-TABLE",
               false);
         ErrorManager.recordOrThrowError(3131,
               "Unable to set attribute ERROR-STRING in widget of type TEMP-TABLE",
               false);
         return;
      }

      this.errorString.assign(errStr);
   }

   /**
    * Obtain the current ERROR-STRING attribute of 4GL temp-table.
    *
    * @return  the current ERROR-STRING of the temp-table
    */
   public character errorString()
   {
      return errorString;
   }

   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      return new character(this.admData);
   }

   /**
     * Set the value of the ADM-DATA attribute.
     * 
     * @param    value
     *           The new value.
     */
   @Override
   public void setADMData(String value)
   {
      this.admData.assign(value);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      this.admData.assign(value);
   }

   /**
    * Returns the state of the ERROR attribute
    *
    * @return    Current state of the ERROR attribute.
    */
   @Override
   public logical isError()
   {
      if (!prepared().booleanValue())
      {
         displayNotPrepared();
      }

      return error;
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   @Override
   public void setError(boolean value)
   {
      setError(new logical(value));
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   @Override
   public void setError(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute ERROR on TEMP-TABLE widget",
               false);
         return;
      }

      if (!prepared().booleanValue())
      {
         displayNotPrepared();
         ErrorManager.recordOrThrowError(3131,
               "Unable to set attribute ERROR in widget of type TEMP-TABLE",
               false);
         return;
      }

      error.assign(value);
   }

   /**
    * Checks whether this temp-table contains any records.
    *
    * @return  true if this temp-table has records.
    */
   @Override
   public logical hasRecords()
   {
      Persistence persistence = PersistenceFactory.getInstance(DatabaseManager.TEMP_TABLE_DB);
      StringBuilder sb = new StringBuilder("select 1 from ");
      sb.append(defaultBuffer().getTable());
      sb.append(" where ");
      sb.append(TemporaryBuffer.MULTIPLEX_FIELD_NAME);
      sb.append(" = ");
      sb.append(defaultBuffer().getMultiplexID());
      try
      {
         ScrollableResults res = persistence.executeSQLQuery(sb.toString());
         if (res.first())
            return new logical(true);
      }
      catch (PersistenceException e)
      {
         ErrorManager.recordOrThrowError(e);
         return new logical(false);
      }

      Iterator<RecordBuffer> iter = BufferManager.get().activeBuffers(persistence, true);
      while(iter.hasNext())
      {
         RecordBuffer buffer = iter.next();
         TempTable tempTable = buffer.getParentTable();
         if (tempTable.equals(this))
            return new logical(true);
      }

      return new logical(false);
   }

   /**
    * Get default buffer for the static temp-table.
    *
    * @return default buffer for the static temp-table.
    */
   RecordBuffer defaultBuffer()
   {
      return RecordBuffer.get((DataModelObject) defaultBufferHandle.getResource());
   }

   /**
    * Display "CREATE-LIKE must be the first method on a newly CREATEd TEMP-TABLE, or a newly
    * CLEARed one" error message.
    */
   protected void displayCreateLikeNotFirst()
   {
      ErrorManager.displayError(9041,
            "CREATE-LIKE must be the first method on a newly CREATEd TEMP-TABLE, " +
                  "or a newly CLEARed one",
            false);
   }

   /**
    * Display "Definitional TEMP-TABLE methods (ADD-NEW/LIKE) may not be called after
    * TEMP-TABLE-PREPARE has been called unless a CLEAR is done first, and may never be called on
    * a DEFINED static table" error message.
    */
   protected void displayPrepared()
   {
      ErrorManager.displayError(9042,
            "Definitional TEMP-TABLE methods (ADD-NEW/LIKE) may not be called after " +
                  "TEMP-TABLE-PREPARE has been called unless a CLEAR is done first, and may never " +
                  "be called on a DEFINED static table",
            false);

   }

   /**
    * Display "TEMP-TABLE is already prepared--ignoring" error message.
    */
   protected void displayPreparedIgnoring()
   {
      ErrorManager.displayError(9043, "TEMP-TABLE is already prepared--ignoring", false);
   }

   /**
    * Display "Unable to assign UNKNOWN value to attribute UNDO on TEMP-TABLE widget" error
    * message.
    */
   protected void displayUnableAssignUnknownUndo()
   {
      ErrorManager.displayError(4083,
            "**Unable to assign UNKNOWN value to attribute UNDO on TEMP-TABLE widget",
            false);
   }

   /**
    * Display "CLEAR method may not be called on a static temp-table object, nor on a dynamic one
    * with records; use buffer:EMPTY-TEMP-TABLE" error message.
    */
   protected void displayCannotClear()
   {
      ErrorManager.displayError(9085,
            "CLEAR method may not be called on a static temp-table object, nor on a dynamic " +
                  "one with records; use buffer:EMPTY-TEMP-TABLE",
            false);
   }

   /**
    * Display "TEMP-TABLE must be PREPARED before referencing NAME/DEFAULT-BUFFER-HANDLE etc"
    * error message.
    */
   protected void displayNotPrepared()
   {
      ErrorManager.displayError(9040,
            "TEMP-TABLE must be PREPARED before referencing NAME/DEFAULT-BUFFER-HANDLE etc",
            false);

   }
}
