/*
** Module   : TableMapper.java
** Abstract : Table and field name to converted DMO and property name (bidirectional).
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20131015 Created initial version.
** 002 CA  20131021 Added legacy index support.
** 003 CA  20131029 mapPermanentDMO must register the mapper for the permanent schema in the
**                  permanentDBs map.
** 004 SVL 20131115 Changes to functions that return DMO class.
** 005 SVL 20140106 Fixed mapping logic for temp tables.
** 006 OM  20140117 Added support for indexed access to fields of the buffer.
** 007 SVL 20140123 Added getPropertyName(Class, int).
** 008 VMN 20140122 Added format(), initial(), columnLabel(), label(), mandatory(),
**                  validateMessage(), validateExpression() support.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.util.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

/**
 * This class performs a bi-directional runtime mapping of DMO implementation classes (permanent 
 * or temporary) to their associated legacy table name. The mapping for a DMO implementation class
 * is held in a {@link LegacyTableInfo} instance.
 * <p>
 * For physical DBs, the mappings for the permanent DBs (connected at server startup) will be 
 * registered at server startup. For permanent DBs connected at application runtime, the mappings
 * will be registered and deregistered at the appropriate DB connection/disconnection times.
 * <p>
 * All the mappings for temporary tables will be registered during application lifetime, as the
 * temp tables get created or dropped.
 * <p>
 * APIs which resolve a legacy name for a converted name may return null, when resolving an index,
 * field or property name; this is to allow checks for synthetic properties or indexes, which do
 * not exist in the legacy code.  In all cases, it is expected that the given table name is for a
 * table which has registered mappings; else, a {@link NullPointerException} will be thrown. 
 * 
 * @param    <K>
 *           The type of the key used by the {@link #p2j progress-to-java} map. For permanent
 *           DMOs, this will be a {@link String}. For temporary DMOs, this will be an 
 *           {@link Integer} value with the <code>UNIQUE-ID</code> identifying the 
 *           {@link TempTable} resource.
 */
public class TableMapper<K>
{
   /** Holds the mappers for all permanent DBs, by their schema name. */
   private static final Map<String, PermanentTableMapper> permanentDBs = 
      new HashMap<String, PermanentTableMapper>();

   /** Context-local mapper for all temporary tables. */
   private static final ContextLocal<TempTableMapper> tempDB = new ContextLocal<TempTableMapper>()
   {
      protected TempTableMapper initialValue()
      {
         return new TempTableMapper();
      };
   };
   
   /** Mapping of DMO implementation classes to their {@link LegacyTableInfo} instance. */
   protected final Map<Class<?>, LegacyTableInfo> j2p = new HashMap<Class<?>, LegacyTableInfo>();
   
   /**
    * Mapping of the legacy tables (identified by a custom key) to their {@link LegacyTableInfo} 
    * instance. */
   protected final Map<K, LegacyTableInfo> p2j = new HashMap<K, LegacyTableInfo>();

   /** The schema name. */
   private final String schema;
   
   /**
    * Create a new mapper for the given schema.
    * 
    * @param    schema
    *           The schema name.
    */
   private TableMapper(String schema)
   {
      this.schema = schema;
   }
   
   /**
    * Get the legacy table name for the DMO referenced by the given buffer.
    * 
    * @param    buffer
    *           A buffer instance.
    * 
    * @return   The legacy table name.
    */
   public static String getLegacyName(Buffer buffer)
   {
      RecordBuffer buf = ((BufferImpl) buffer).buffer();
      if (buf.isTemporary())
      {
         TempTable tt = buf.getParentTable();
         String name = tt.name().getValue();
         if (name != null)
         {
            // return the name of the parent table
            return name;
         }
         else
         {
            // name not yet assigned to a static temp table, get it from the mapping
            return getLegacyName(tt);
         }
      }
      else
      {
         return getLegacyName(buf.getDMOImplementationClass());
      }
   }

   /**
    * Get the COLUMN-LABEL field attribute for the property associated with the DMO referenced
    * by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   COLUMN-LABEL field attribute or <code>null</code> if there is no such field.
    */
   public static String getLegacyFieldColumnLabel(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      return legacyFieldInfo == null ? null : legacyFieldInfo.columnLabel;
   }

   /**
    * Set the COLUMN-LABEL field attribute for the property associated with the DMO referenced
    * by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * @param    columnLabel
    *           COLUMN-LABEL field attribute.
    */
   public static void setLegacyFieldColumnLabel(Buffer buffer,
                                                String property,
                                                String columnLabel)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      legacyFieldInfo.columnLabel = columnLabel;
   }

   /**
    * Get the LABEL field attribute for the property associated with the DMO referenced
    * by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   LABEL field attribute or <code>null</code> if there is no such field.
    */
   public static String getLegacyFieldLabel(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      return legacyFieldInfo == null ? null : legacyFieldInfo.label;
   }

   /**
    * Set the LABEL field attribute for the property associated with the DMO referenced
    * by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * @param    label
    *           LABEL field attribute.
    */
   public static void setLegacyFieldLabel(Buffer buffer, String property, String label)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      legacyFieldInfo.label = label;
   }

   /**
    * Get the id of field in the sequence of the fields of the buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   the id of field in the sequence of the fields of the buffer.
    */
   public static integer getLegacyFieldId(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      return new integer(legacyFieldInfo == null ? null : legacyFieldInfo.fieldId);
   }

   /**
    * Get the LITERAL-QUESTION field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   LITERAL-QUESTION field attribute or <code>null</code> if there is no such field.
    */
   public static logical getLegacyFieldLiteralQuestion(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      return new logical(legacyFieldInfo == null ? null : legacyFieldInfo.literalQuestion);
   }

   /**
    * Set the LITERAL-QUESTION field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * @param    literalQuestion
    *           LITERAL-QUESTION field attribute.
    */
   public static void setLegacyFieldLiteralQuestion(Buffer buffer,
                                                    String property,
                                                    logical literalQuestion)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      legacyFieldInfo.literalQuestion = literalQuestion.booleanValue();
   }

   /**
    * Get the MANDATORY field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   MANDATORY field attribute or <code>null</code> if there is no such field.
    */
   public static logical isLegacyFieldMandatory(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      return new logical(legacyFieldInfo == null ? null : legacyFieldInfo.mandatory);
   }

   /**
    * Get the VALIDATE-EXPRESSION field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   VALIDATE-EXPRESSION field attribute or <code>null</code> if there is no such
    *           field.
    */
   public static String getLegacyFieldValidateExpression(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      return legacyFieldInfo == null ? null : legacyFieldInfo.validateExpression;
   }

   /**
    * Set the VALIDATE-EXPRESSION field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * @param    validateExpression
    *           VALIDATE-EXPRESSION field attribute.
    */
   public static void setLegacyFieldValidateExpression(Buffer buffer,
                                                    String property,
                                                    String validateExpression)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      legacyFieldInfo.validateExpression = validateExpression;
   }

   /**
    * Get the VALIDATE-MESSAGE field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   VALIDATE-MESSAGE field attribute or <code>null</code> if there is no
    *           such field.
    */
   public static String getLegacyFieldValidateMessage(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      return legacyFieldInfo == null ? null : legacyFieldInfo.validateMessage;
   }

   /**
    * Set the VALIDATE-MESSAGE field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * @param    validateMessage
    *           VALIDATE-MESSAGE field attribute.
    */
   public static void setLegacyFieldValidateMessage(Buffer buffer,
                                                    String property,
                                                    String validateMessage)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      legacyFieldInfo.validateMessage = validateMessage;
   }

   /**
    * Get the FORMAT field attribute for the property associated with the DMO referenced by the
    * given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   FORMAT field attribute or <code>null</code> if there is no such field.
    */
   public static String getLegacyFieldFormat(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      return legacyFieldInfo == null ? null : legacyFieldInfo.format;
   }

   /**
    * Set the FORMAT field attribute for the property associated with the DMO referenced by the
    * given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * @param    format
    *           FORMAT field attribute.
    */
   public static void setLegacyFieldFormat(Buffer buffer, String property, String format)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      legacyFieldInfo.format = format;
   }

   /**
    * Get the INITIAL field attribute for the property associated with the DMO referenced by the
    * given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   INITIAL field attribute or <code>null</code> if there is no such field.
    */
   public static String getLegacyFieldInitial(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      return legacyFieldInfo == null ? null : legacyFieldInfo.initial;
   }

   /**
    * Get the legacy field name for the property associated with the DMO referenced by the given
    * buffer.
    * 
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * 
    * @return   The legacy field name or <code>null</code> if there is no such field.
    */
   public static String getLegacyFieldName(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      return legacyFieldInfo == null ? null : legacyFieldInfo.legacyName;
   }

   /**
    * Get the legacy field info for the property associated with the DMO referenced by the given
    * buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   The legacy field info or <code>null</code> if there is no such field.
    */
   private static LegacyFieldInfo getLegacyFieldInfo(Buffer buffer, String property)
   {
      RecordBuffer buf = ((BufferImpl) buffer).buffer();
      if (buf.isTemporary())
      {
         return getLegacyFieldInfo((TempTable) buffer.tableHandle().get(), property);
      }
      else
      {
         return getLegacyFieldInfo(buf.getDMOImplementationClass(), property);
      }
   }

   /**
    * Get the legacy index name of a converted index, defined by the backing table of the 
    * specified buffer.
    * 
    * @param    buffer
    *           A buffer instance.
    * @param    idx
    *           The converted index name.
    *           
    * @return   The legacy index name or <code>null</code> if there is no such index defined.
    */
   public static String getLegacyIndexName(Buffer buffer, String idx)
   {
      RecordBuffer buf = ((BufferImpl) buffer).buffer();
      if (buf.isTemporary())
      {
         return getLegacyIndexName((TempTable) buffer.tableHandle().get(), idx);
      }
      else
      {
         return getLegacyIndexName(buf.getDMOImplementationClass(), idx);
      }
   }

   /**
    * Get the legacy table name for the given DMO, which is part of a permanent DB.
    * 
    * @param    dmoClass
    *           The DMO implementation class for a permanent DMO.
    *           
    * @return   The legacy table name.
    * 
    * @throws   IllegalStateException
    *           If a DMO belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} temp schema
    *           is passed as a parameter; for these cases, {@link #getLegacyName(TempTable)}
    *           or {@link #getLegacyName(Buffer)} must be used.
    */
   public static String getLegacyName(Class<?> dmoClass)
   {
      String schema = DatabaseManager.getSchemaByClass(dmoClass);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg =
            "Legacy table name for a temporary DMO can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }

      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);

         return mapper.legacyName(dmoClass);
      }
   }
   
   /**
    * Get the legacy field name for the given property, which is defined by a DMO part of a 
    * permanent DB.
    * 
    * @param    dmoClass
    *           The DMO implementation class for a permanent DMO.
    * @param    property
    *           The DMO property name.
    *           
    * @return   The legacy field name or <code>null</code> if there is no such field.
    * 
    * @throws   IllegalStateException
    *           If a DMO belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} temp schema
    *           is passed as a parameter; for these cases, {@link #getLegacyFieldName(TempTable, 
    *           String)} or {@link #getLegacyFieldName(Buffer, String)} must be used.
    */
   public static String getLegacyFieldName(Class<?> dmoClass, String property)
   {
      String schema = DatabaseManager.getSchemaByClass(dmoClass);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg =
            "Legacy field names for a property of a temporary DMO can not be resolved using " +
            "this API!";
         throw new IllegalArgumentException(msg);
      }

      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         
         return mapper.legacyFieldName(dmoClass, property);
      }
   }

   /**
    * Get the legacy field for the given property, which is defined by a DMO part of a
    * permanent DB.
    *
    * @param    dmoClass
    *           The DMO implementation class for a permanent DMO.
    * @param    property
    *           The DMO property name.
    *
    * @return   The legacy field or <code>null</code> if there is no such field.
    *
    * @throws   IllegalStateException
    *           If a DMO belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} temp schema
    *           is passed as a parameter; for these cases, {@link #getLegacyFieldInfo(TempTable,
    *           String)} or {@link #getLegacyFieldInfo(Buffer, String)} must be used.
    */
   private static LegacyFieldInfo getLegacyFieldInfo(Class<?> dmoClass, String property)
   {
      String schema = DatabaseManager.getSchemaByClass(dmoClass);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg =
                 "Legacy field names for a property of a temporary DMO can not be resolved using " +
                         "this API!";
         throw new IllegalArgumentException(msg);
      }

      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);

         return mapper.legacyFieldInfo(dmoClass, property);
      }
   }

   /**
    * Get the legacy name of a converted index, which is defined by the backing table for the 
    * specified buffer, part of a permanent DB.
    * 
    * @param    dmoClass
    *           The DMO implementation class for a permanent DMO.
    * @param    idx
    *           The converted index name.
    *           
    * @return   The legacy index name or <code>null</code> if there is no such index defined.
    * 
    * @throws   IllegalStateException
    *           If a DMO belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} temp schema
    *           is passed as a parameter; for these cases, {@link #getLegacyFieldName(TempTable, 
    *           String)} or {@link #getLegacyFieldName(Buffer, String)} must be used.
    */
   public static String getLegacyIndexName(Class<?> dmoClass, String idx)
   {
      String schema = DatabaseManager.getSchemaByClass(dmoClass);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg =
            "Legacy index names for a temporary DMO can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }

      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         
         return mapper.legacyIndexName(dmoClass, idx);
      }
   }
   
   /**
    * Get the legacy table name for the given {@link TempTable} resource.
    * 
    * @param    tt
    *           The {@link TempTable} resource.
    *           
    * @return   The legacy table name.
    */
   public static String getLegacyName(TempTable tt)
   {
      TempTableMapper mapper = locateTemp();
      return mapper.legacyName(tt);
   }
   
   /**
    * Get the legacy field name for the given {@link TempTable} resource and property.
    * 
    * @param    tt
    *           The {@link TempTable} resource.
    * @param    property
    *           The DMO property name.
    * 
    * @return   The legacy field name or <code>null</code> if there is no such field.
    */
   public static String getLegacyFieldName(TempTable tt, String property)
   {
      TempTableMapper mapper = locateTemp();
      return mapper.legacyFieldInfo(tt, property).legacyName;
   }

   /**
    * Get the legacy field for the given {@link TempTable} resource and property.
    *
    * @param    tt
    *           The {@link TempTable} resource.
    * @param    property
    *           The DMO property name.
    *
    * @return   The legacy field or <code>null</code> if there is no such field.
    */
   public static LegacyFieldInfo getLegacyFieldInfo(TempTable tt, String property)
   {
      TempTableMapper mapper = locateTemp();
      return mapper.legacyFieldInfo(tt, property);
   }

   /**
    * Get the legacy index name for the given {@link TempTable} resource and converted index name.
    * 
    * @param    tt
    *           The {@link TempTable} resource.
    * @param    idx
    *           The converted index name.
    *           
    * @return   The legacy index name or <code>null</code> if there is no such index defined.
    */
   public static String getLegacyIndexName(TempTable tt, String idx)
   {
      TempTableMapper mapper = locateTemp();
      return mapper.legacyIndexName(tt, idx);
   }

   /**
    * Get the DMO implementation class for the specified {@link TempTable} resource.
    * 
    * @param    table
    *           The {@link TempTable} resource.
    *           
    * @return   The DMO implementation class.
    */
   public static Class<?> getDMOClass(TempTable table)
   {
      // no reason to search the mappings, get it from the persistence info
      return table.getDMOClass();
   }

   /**
    * Resolve the DMO property name for a legacy field name belonging to the specified 
    * {@link TempTable} resource.
    * 
    * @param    table
    *           The {@link TempTable} resource.
    * @param    field
    *           The legacy field name.
    *           
    * @return   The DMO property name associated with the legacy field or <code> if there is no 
    *           such field.
    */
   public static String getPropertyName(TempTable table, String field)
   {
      TempTableMapper mapper = locateTemp();
      return mapper.getPropertyName(table.getUniqueID().toJavaIntegerType(), field);
   }

   /**
    * Resolve the DMO property name for a field index name belonging to the specified
    * {@link TempTable} resource.
    *
    * @param    table
    *           The {@link TempTable} resource.
    * @param    fieldId
    *           The field index in table declaring sequence.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    */
   public static String getPropertyName(TempTable table, int fieldId)
   {
      TempTableMapper mapper = locateTemp();
      return mapper.getPropertyName(table.getUniqueID().toJavaIntegerType(), fieldId);
   }

   /**
    * Resolve the converted index name for a legacy index belonging to the temp-table defined
    * by the {@link TempTable} resource.
    * 
    * @param    table
    *           The {@link TempTable} resource.
    * @param    idx
    *           The legacy index name.
    *           
    * @return   The converted index name or <code>null</code> if there is no such index.
    */
   public static String getIndexName(TempTable table, String idx)
   {
      TempTableMapper mapper = locateTemp();
      return mapper.getIndexName(table.getUniqueID().toJavaIntegerType(), idx);
   }
   
   /**
    * Get the DMO implementation class for the legacy table name (part of the specified schema).
    * 
    * @param    schema
    *           The schema name.
    * @param    table
    *           The legacy table name.
    *           
    * @return   The DMO implementation class or <code>null</code> if there is no table with the
    *           given name.
    * 
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static Class<?> getDMOClass(String schema, String table)
   {
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg = "DMO for a temporary table can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }

      PermanentTableMapper mapper = locatePerm(schema);
      return mapper.getDMOClass(table.toLowerCase());
   }
   
   /**
    * Get the DMO property name for a legacy field name belonging to the specified legacy table.
    * 
    * @param    schema
    *           The schema name.
    * @param    table
    *           The legacy table name.
    * @param    field
    *           The legacy field name.
    *           
    * @return   The DMO property name associated with the legacy field or <code> if there is no 
    *           such field.
    * 
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static String getPropertyName(String schema, String table, String field)
   {
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg = "Properties for temporary DMOs can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }

      PermanentTableMapper mapper = locatePerm(schema);
      return mapper.getPropertyName(table.toLowerCase(), field);
   }

   /**
    * Get the DMO property name for a field index belonging to the specified legacy table.
    *
    * @param    schema
    *           The schema name.
    * @param    table
    *           The legacy table name.
    * @param    fieldId
    *           The field index.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    *
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static String getPropertyName(String schema, String table, int fieldId)
   {
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg = "Properties for temporary DMOs can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }

      PermanentTableMapper mapper = locatePerm(schema);
      return mapper.getPropertyName(table.toLowerCase(), fieldId);
   }

   /**
    * Get the DMO property name for a field index belonging to the specified legacy table.
    *
    * @param    dmoClass
    *           The DMO implementation class for a permanent DMO.
    * @param    fieldId
    *           The field index.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    *
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static String getPropertyName(Class<?> dmoClass, int fieldId)
   {
      String schema = DatabaseManager.getSchemaByClass(dmoClass);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg = "Properties for temporary DMOs can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }

      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         return mapper.propertyName(dmoClass, fieldId);
      }
   }

   /**
    * Resolve the converted index name for a legacy index belonging to the specified legacy table.
    * 
    * @param    schema
    *           The schema name.
    * @param    table
    *           The legacy table name.
    * @param    idx
    *           The legacy index name.
    *           
    * @return   The converted index name or <code>null</code> if there is no such index.
    * 
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static String getIndexName(String schema, String table, String idx)
   {
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg = "Properties for temporary DMOs can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }

      PermanentTableMapper mapper = locatePerm(schema);
      return mapper.getIndexName(table.toLowerCase(), idx);
   }

   /**
    * Add the legacy mappings for the specified DMO.
    * 
    * @param    schema
    *           The schema to which the DMO belongs.
    * @param    dmoClass
    *           The MDO implementation class.
    */
   static void mapPermanentDMO(String schema, Class<?> dmoClass)
   {
      // - mapping is kept by schema
      // - permamentDBs are kept in a global map
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         if (mapper == null)
         {
            mapper = new PermanentTableMapper(schema);
            permanentDBs.put(schema,  mapper);
         }
         
         mapper.mapClass(dmoClass);
      }
   }

   /**
    * Remove the mappings for the specified schema.
    * 
    * @param    schema
    *           The schema name.
    */
   static void removePermanentSchema(String schema)
   {
      synchronized (permanentDBs)
      {
         permanentDBs.remove(schema);
      }
   }

   /**
    * Remove all the temporary tables for this context.
    */
   static void removeTemporaryTables()
   {
      locateTemp().clear();
   }
   
   /**
    * Add the legacy mappings for the given temporary table.
    * 
    * @param    table
    *           The {@link TempTable} resource.
    */
   static void mapTemporaryTable(TempTable table)
   {
      if (table == null)
      {
         // TODO: this should throw an NPE, but for know leave it like this, to pass regression
         // testing
         return;
      }
      
      locateTemp().mapClass(table);
   }

   /**
    * Remove the mappings for the given temporary table.
    * 
    * @param    table
    *           The {@link TempTable} resource to be removed.
    */
   static void removeTemporaryTable(TempTable table)
   {
      locateTemp().removeClass(table);
   }
   
   /**
    * Locate the context-local {@link TempTableMapper}.
    * 
    * @return   See above.
    */
   private static TempTableMapper locateTemp()
   {
      return tempDB.get();
   }
   
   /**
    * Locate the {@link PermanentTableMapper} instance for the given schema.
    * 
    * @param    schema
    *           The schema name.
    *           
    * @return   See above.
    */
   private static PermanentTableMapper locatePerm(String schema)
   {
      return permanentDBs.get(schema);
   }

   /**
    * Get the legacy table name associated with the given DMO implemenation class.
    * 
    * @param    dmoClass
    *           The DMO implementation class.
    *           
    * @return   The legacy table name.
    */
   protected String legacyName(Class<?> dmoClass)
   {
      LegacyTableInfo tableInfo = j2p.get(dmoClass);
      
      return tableInfo.legacyName;
   }

   /**
    * Get the legacy field name for given property defined by the specified DMO.
    * 
    * @param    dmoClass
    *           The DMO implementation class.
    * @param    property
    *           The DMO property name.
    *           
    * @return   The legacy field name or <code>null</code> if there is no such field.
    */
   protected String legacyFieldName(Class<?> dmoClass, String property)
   {
      LegacyTableInfo tableInfo = j2p.get(dmoClass);
      LegacyFieldInfo fieldInfo = tableInfo.j2p.get(property);
      
      return (fieldInfo == null ? null : fieldInfo.legacyName);
   }

   /**
    * Get the DMO property name for a field index belonging to the specified legacy table.
    *
    * @param    dmoClass
    *           The DMO implementation class for a permanent DMO.
    * @param    fieldId
    *           The field index.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    */
   protected String propertyName(Class<?> dmoClass, int fieldId)
   {
      LegacyTableInfo tableInfo = j2p.get(dmoClass);
      LegacyFieldInfo fieldInfo = tableInfo.ip2j.get(fieldId);

      return (fieldInfo == null ? null : fieldInfo.legacyName);
   }
   
   /**
    * TODO: add javadoc
    */
   protected LegacyFieldInfo legacyFieldInfo(Class<?> dmoClass, String property)
   {
      LegacyTableInfo tableInfo = j2p.get(dmoClass);
      return tableInfo.j2p.get(property);
   }
   
   /**
    * Get the legacy index name of converted index, which is defined by the backing table for the 
    * specified DMO implementation class.
    * 
    * @param    dmoClass
    *           The DMO implementation class.
    * @param    idx
    *           The converted index name.
    *           
    * @return   The legacy index name or <code>null</code> if there is no such index defined.
    */
   protected String legacyIndexName(Class<?> dmoClass, String idx)
   {
      LegacyTableInfo tableInfo = j2p.get(dmoClass);
      LegacyIndexInfo indexInfo = tableInfo.idxj2p.get(idx);
      
      return (indexInfo == null ? null : indexInfo.legacyName);
   }
   
   /**
    * Map the specified DMO implementation class using the specified <code>legacyKey</code>.
    * <p>
    * Before calling this method, make sure the passed DMO has a {@link LegacyTable} annotation.
    * 
    * @param    legacyKey
    *           The legacy key to identify this DMO implementation class.
    * @param    dmoClass
    *           The DMO implementation class.
    */
   protected void mapClass(K legacyKey, Class<?> dmoClass)
   {
      if (j2p.containsKey(dmoClass))
      {
         // already mapped
         return;
      }
      
      LegacyTableInfo tableInfo = new LegacyTableInfo(dmoClass);
      
      p2j.put(legacyKey, tableInfo);
      j2p.put(tableInfo.dmoClass, tableInfo);
   }
   
   /**
    * Remove the mappings for the specified DMO implementation class.
    * 
    * @param    dmoClass
    *           The DMO implementation class.
    */
   protected void removeClass(Class<?> dmoClass)
   {
      LegacyTableInfo info = j2p.get(dmoClass);
      
      if (info == null)
      {
         // ignore if not found
         return;
      }
      
      j2p.remove(info.dmoClass);
      p2j.remove(info.legacyName);
   }
   
   /**
    * Get the DMO class for the legacy table identified with the specified <code>legacyKey</code>.
    * 
    * @param    legacyKey
    *           The key used to identify the {@link LegacyTableInfo}.
    * 
    * @return   The DMO implementation class or <code>null</code> if there is no class
    *           corresponding the given key.
    */
   protected Class<?> getDMOClass(K legacyKey)
   {
      LegacyTableInfo tableInfo = p2j.get(legacyKey);
      
      return tableInfo == null ? null : tableInfo.dmoClass;
   }

   /**
    * Get the property name for the specified field. The legacy table info is determined using
    * the specified <code>legacyKey</code>.
    * 
    * @param    legacyKey
    *           The key used to identify the {@link LegacyTableInfo}.
    * @param    field
    *           The legacy field name.
    *           
    * @return   The DMO property name associated with the legacy field or <code> if there is no 
    *           such field.
    */
   protected String getPropertyName(K legacyKey, String field)
   {
      LegacyTableInfo tableInfo = p2j.get(legacyKey);
      LegacyFieldInfo fieldInfo = tableInfo.p2j.get(field.toLowerCase());
      
      return (fieldInfo == null ? null : fieldInfo.javaName);
   }
   /**
    * Get the property name for the specified field index. The legacy table info is determined
    * using the specified <code>legacyKey</code>.
    *
    * @param    legacyKey
    *           The key used to identify the {@link LegacyTableInfo}.
    * @param    fieldId
    *           The index of the field in the buffer's sequence.
    *
    * @return   The DMO property name associated with the field index or <code>null</code> if
    *           there is no such field.
    */
   protected String getPropertyName(K legacyKey, int fieldId)
   {
      LegacyTableInfo tableInfo = p2j.get(legacyKey);
      LegacyFieldInfo fieldInfo = tableInfo.ip2j.get(fieldId);

      return (fieldInfo == null ? null : fieldInfo.javaName);
   }

   /**
    * Get the converted index name for the specified legacy index. The legacy table info is 
    * determined using the specified <code>legacyKey</code>.
    * 
    * @param    legacyKey
    *           The key used to identify the {@link LegacyTableInfo}.
    * @param    idx
    *           The legacy index name.
    *           
    * @return   The converted index name or <code>null</code> if there is no such index.
    */
   protected String getIndexName(K legacyKey, String idx)
   {
      LegacyTableInfo tableInfo = p2j.get(legacyKey);
      LegacyIndexInfo indexInfo = tableInfo.idxp2j.get(idx.toLowerCase());
      
      return (indexInfo == null ? null : indexInfo.convertedName);
   }
   
   /**
    * Clear all the mappings.
    */
   protected void clear()
   {
      j2p.clear();
      p2j.clear();
   }
   
   /**
    * Mapper for permanent tables.
    */
   private static class PermanentTableMapper
   extends TableMapper<String>
   {
      /**
       * Default c'tor.
       * 
       * @param    schema
       *           The schema name.
       */
      public PermanentTableMapper(String schema)
      {
         super(schema);
      }

      /**
       * Register the mappings for the specified permanent DMO.
       * 
       * @param    dmoClass
       *           The DMO implementation class.
       * 
       * @throws   NullPointerException
       *           If the DMO class does not have a {@link LegacyTable} annotation.
       */
      protected void mapClass(Class<?> dmoClass)
      {
         if (!dmoClass.isAnnotationPresent(LegacyTable.class))
         {
            final String msg = "DMO class %s must have a LegacyTable annotation!"; 
            throw new NullPointerException(String.format(msg, dmoClass.getName()));
         }
         
         LegacyTable table = dmoClass.getAnnotation(LegacyTable.class);
         
         mapClass(table.name().toLowerCase(), dmoClass);
      }
   }
   
   /**
    * Mapper for temporary tables.
    */
   private static class TempTableMapper
   extends TableMapper<Integer>
   {
      /**
       * Default c'tor.
       */
      public TempTableMapper()
      {
         super(DatabaseManager.TEMP_TABLE_SCHEMA);
      }

      /**
       * Register the mappings for the specified temp table.
       * 
       * @param    tt
       *           The temp-table which needs to be mapped.
       *           
       * @throws   NullPointerException
       *           If the associated DMO class does not have a {@link LegacyTable} annotation.
       */
      protected void mapClass(TempTable tt)
      {
         Class<?> dmoClass = getDMOClass(tt);
         
         if (!dmoClass.isAnnotationPresent(LegacyTable.class))
         {
            final String msg = "DMO class %s must have a LegacyTable annotation!"; 
            throw new NullPointerException(String.format(msg, dmoClass.getName()));
         }

         Integer key = getLegacyKey(tt);

         if (p2j.containsKey(key))
         {
            // already mapped
            return;
         }

         LegacyTableInfo tableInfo = new LegacyTableInfo(dmoClass);
         p2j.put(key, tableInfo);
      }

      /**
       * Remove the mappings for the specified temp table.
       *
       * @param    tt
       *           Temp table object.
       */
      protected void removeClass(TempTable tt)
      {
         Integer key = getLegacyKey(tt);
         p2j.remove(key);
      }

      /**
       * Get the legacy table name associated with the given temp table.
       *
       * @param    tt
       *           Temp table object.
       *
       * @return   The legacy table name.
       */
      protected String legacyName(TempTable tt)
      {
         Integer key = getLegacyKey(tt);
         LegacyTableInfo info = p2j.get(key);
         return info.legacyName;
      }

      /**
       * Get the legacy field name for given property defined by the specified temp table.
       *
       * @param    tt
       *           Temp table object.
       * @param    property
       *           The DMO property name.
       *
       * @return   The legacy field name or <code>null</code> if there is no such field.
       */
      private LegacyFieldInfo legacyFieldInfo(TempTable tt, String property)
      {
         Integer key = getLegacyKey(tt);
         LegacyTableInfo tableInfo = p2j.get(key);
         return  tableInfo.j2p.get(property);
      }

      /**
       * Get the legacy index name of converted index, which is defined by the specified backing
       * temp table.
       *
       * @param    tt
       *           Temp table object.
       * @param    idx
       *           The converted index name.
       *
       * @return   The legacy index name or <code>null</code> if there is no such index defined.
       */
      protected String legacyIndexName(TempTable tt, String idx)
      {
         Integer key = getLegacyKey(tt);
         LegacyTableInfo tableInfo = p2j.get(key);
         LegacyIndexInfo indexInfo = tableInfo.idxj2p.get(idx);
         return indexInfo.legacyName;
      }

      /**
       * Get legacy key for the specified temp table.
       *
       * @param  tt
       *         Temp table object.
       *
       * @return legacy key for the specified temp table.
       */
      protected Integer getLegacyKey(TempTable tt)
      {
         return tt.getUniqueID().toJavaIntegerType();
      }
   }
   
   /**
    * A container with the legacy info for a table.
    */
   private static class LegacyTableInfo
   {
      /** Field info kept by legacy name in progress. */ 
      private final Map<String, LegacyFieldInfo> p2j = new HashMap<>();
      
      /**
       * Field info kept by legacy buffer field sequence index in progress.
       * This is needed at runtime by the buffer-field attribute.
       */
      private final Map<Integer, LegacyFieldInfo> ip2j = new HashMap<>();

      /** Field info kept by associated DMO property name. */
      private final Map<String, LegacyFieldInfo> j2p = new HashMap<>();
      
      /** Index info kept by legacy index name. */
      private final Map<String, LegacyIndexInfo> idxp2j = new HashMap<>();
      
      /** Index info kept by converted index name. */
      private final Map<String, LegacyIndexInfo> idxj2p = new HashMap<>();

      /** The legacy table name. */
      private final String legacyName;
      
      /** The DMO class. */
      private final Class<?> dmoClass;
      
      /**
       * Initialize the legacy info for a table by reading the {@link LegacyTable} and 
       * {@link LegacyField} annotation from the DMO implementation class.
       * 
       * @param    dmoClass
       *           The DMO class from where the annotations will be read.
       * 
       * @see      #loadFields(Class)
       */
      public LegacyTableInfo(Class<?> dmoClass)
      {
         LegacyTable ltable = dmoClass.getAnnotation(LegacyTable.class);
         
         this.legacyName = ltable.name();
         this.dmoClass = dmoClass;
         
         loadFields(dmoClass);
         
         Class<?>[] classMembers = dmoClass.getDeclaredClasses();
         for (Class<?> cls : classMembers)
         {
            loadFields(cls);
         }
         
         LegacyIndexes lindexes = dmoClass.getAnnotation(LegacyIndexes.class);
         if (lindexes != null)
         {
            for (LegacyIndex lidx : lindexes.value())
            {
               LegacyIndexInfo idxInfo = new LegacyIndexInfo(lidx.name(), lidx.converted());
               
               this.idxp2j.put(idxInfo.legacyName.toLowerCase(), idxInfo);
               this.idxj2p.put(idxInfo.convertedName, idxInfo);
            }
         }
      }
      
      /**
       * Load all fields from the specified class and add their mappings.
       * 
       * @param    cls
       *           The DMO class or a composite class with the extent fields.
       * 
       * @throws   NullPointerException
       *           If a DMO property does not have a {@link LegacyField} annotation.
       */
      private void loadFields(Class<?> cls)
      {
         Field[] fields = cls.getDeclaredFields();
         for (Field field : fields)
         {
            if (!BaseDataType.class.isAssignableFrom(field.getType()))
            {
               // ignore P2J-specific DMO properties. all legacy DMO properties have BaseDataType
               // subclass as their type
               continue;
            }
            
            LegacyField lfield = field.getAnnotation(LegacyField.class);
            
            if (lfield == null)
            {
               final String msg = "Property %s for DMO %s must have a LegacyField annotation!"; 
               throw new NullPointerException(
                  String.format(msg, field.getName(), dmoClass.getName()));
            }

            String columnLabel = lfield.columnLabel();
            String pname = lfield.name();
            String jname = field.getName();
            String format = lfield.format();
            String initial = lfield.initial();
            String label = lfield.label();
            boolean mandatory = lfield.mandatory();
            String validateMessage = lfield.validateMessage();
            String validateExpression = lfield.validateExpression();

            Integer fieldId = lfield.fieldId();

            LegacyFieldInfo fieldInfo = new LegacyFieldInfo(columnLabel,
                                                            fieldId,
                                                            format,
                                                            initial,
                                                            jname,
                                                            label,
                                                            pname,
                                                            mandatory,
                                                            validateMessage,
                                                            validateExpression);
            
            p2j.put(pname.toLowerCase(), fieldInfo);
            j2p.put(jname, fieldInfo);
            ip2j.put(fieldId, fieldInfo);
         }
      }
   }
   
   /**
    * A container with a index's legacy and converted name.
    */
   private static class LegacyIndexInfo
   {
      /** The legacy name of this index. */
      private final String legacyName;
      
      /** The converted name of this index. */
      private final String convertedName;
      
      /**
       * Create a new index information container, holding the legacy and converted names.
       * 
       * @param    legacyName
       *           The legacy name of this index.
       * @param    convertedName
       *           The converted name of this index.
       */
      public LegacyIndexInfo(String legacyName, String convertedName)
      {
         this.legacyName = legacyName;
         this.convertedName = convertedName;
      }
   }
   
   /**
    * A container with a field's legacy and DMO property name and field attributes.
    */
   private static class LegacyFieldInfo
   {
      /** COLUMN-LABEL attribute of this field. */
      private String columnLabel;

      /**
       * The id of this field in the sequence of the fields of the buffer. It is needed by the
       * dynamic <code>bufferField(int)</code> attribute of the <code>BufferImpl</code> for
       * emulating the <code>BUFFER-FIELD(integer)</code> buffer handle attribute.
       */
      private final Integer fieldId;

      /** FORMAT attribute of this field. */
      private String format;

      /** INITIAL attribute of this field. */
      private final String initial;

      /** The associated DMO property name. */
      private final String javaName;

      /** LABEL attribute of this field. */
      private String label;

      /** The legacy name of this field. */
      private final String legacyName;

      /** LITERAL-QUESTION attribute of this field. */
      private boolean literalQuestion = false;

      /** MANDATORY attribute of this field. */
      public final boolean mandatory;

      /** VALIDATE-MESSAGE attribute of this field. */
      private String validateMessage;

      /** VALIDATE-EXPRESSION attribute of this field. */
      private String validateExpression;

      /**
       * Create a new field information container, holding the legacy and converted names and
       * field attributes.
       *
       * @param    columnLabel
       *           COLUMN-LABEL attribute.
       * @param    fieldId
       *           The id of field in the sequence of the fields of the buffer.
       * @param    format
       *           FORMAT attribute.
       * @param    initial
       *           INITIAL attribute.
       * @param    javaName
       *           The field's associated DMO property name.
       * @param    label
       *           LABEL attribute.
       * @param    legacyName
       *           The field's legacy name.
       * @param    mandatory
       *           MANDATORY attribute.
       * @param    validateMessage
       *           VALIDATE-MESSAGE attribute.
       * @param    validateExpression
       *           VALIDATE-EXPRESSION attribute.
       */
      public LegacyFieldInfo(String columnLabel,
                             Integer fieldId,
                             String format,
                             String initial,
                             String javaName,
                             String label,
                             String legacyName,
                             boolean mandatory,
                             String validateMessage,
                             String validateExpression)
      {
         this.columnLabel = columnLabel;
         this.fieldId = fieldId;
         this.format = format;
         this.initial = initial;
         this.javaName = javaName;
         this.label = label;
         this.legacyName = legacyName;
         this.mandatory = mandatory;
         this.validateMessage = validateMessage;
         this.validateExpression = validateExpression;
      }
   }
}
